import { querySelector, elAddCS, ElStyle, elAddCS2 } from '../../utils/query-el';
import { generateId } from '../../utils/nanoid';
import { modalStore } from './store';
import { InitModalEvent } from './event';
import { ObjCss } from '../../utils/css';
export type KV = {
  [key: string]: any;
};
export type ModalOpts<
  T = {
    [key: string]: any;
  },
  U = {
    [key: string]: any;
  },
> = {
  root?: HTMLDivElement | string;
  id?: string;
  mask?: boolean;
  maskClassName?: string;
  maskStyle?: ElStyle;
  maskClose?: boolean;

  contentClassName?: string;
  contentStyle?: ElStyle;
  destroyOnClose?: boolean; // 关闭，把Element移动到cacheFragment中
  hideOnClose?: boolean; // 关闭后是否销毁，设置display:none
  open?: boolean;

  onClose?: () => void;
  defaultStyle?: DefaultStyle<U>;
} & T;
export type DefaultStyle<T> = {
  defaultContentStyle?: ObjCss;
  defaultMaskStyle?: ObjCss;
} & T;

export class Modal<T = any, U = KV> {
  static rootClassName = '#ui-modal-list';
  root: HTMLDivElement;
  id: string;
  modalElement?: HTMLDivElement;
  Element?: HTMLDivElement;
  mask?: boolean;
  maskClassName?: string;
  maskStyle?: ElStyle;
  // 点击mask是否关闭
  maskClose?: boolean;

  contentClassName?: string;
  contentStyle?: ElStyle;

  destroyOnClose?: boolean;
  hideOnClose?: boolean;

  open?: boolean;
  isUse = true;
  // 通知关闭
  onClose?: (e: any) => void;
  cacheFragment?: DocumentFragment;

  defaultStyle?: DefaultStyle<U>;
  constructor(opts: ModalOpts<T, U>) {
    this.root = this.initRoot(opts.root);
    InitModalEvent(this.root);
    this.id = opts.id || generateId();
    this.mask = opts.mask ?? true;
    this.maskClassName = opts.maskClassName;
    this.maskStyle = opts.maskStyle;
    this.maskClose = opts.maskClose ?? true;
    this.contentClassName = opts.contentClassName;
    this.contentStyle = opts.contentStyle;
    this.destroyOnClose = opts.destroyOnClose ?? true;
    this.hideOnClose = opts.hideOnClose ?? true;
    if (!this.destroyOnClose && !this.hideOnClose) {
      this.destroyOnClose = true; // 必须要有一个为true
      console.warn('destroyOnClose Or hideOnClose must one is true');
    }
    this.cacheFragment = new DocumentFragment();
    this.defaultStyle = opts.defaultStyle || ({} as DefaultStyle<U>);
    this.open = opts.open ?? true;
    this.onClose = opts.onClose;
  }
  protected initRoot(root: ModalOpts['root']) {
    let _root = querySelector(root);
    if (!_root) {
      // 查询ui-modal元素,不存在则创建一个ui-modal元素并添加到body上
      const queryRoot = document.querySelector('#ui-modal-list') as HTMLDivElement;
      if (queryRoot) {
        _root = queryRoot;
        _root.classList.add('ui-modal-root');
        return _root;
      }
      _root = document.createElement('div');
      _root.id = 'ui-modal-list';
      _root.classList.add('ui-modal-root');
      document.body.appendChild(_root);
      return _root;
    }
    _root.classList.add('ui-custom-modal', 'ui-modal-root');

    if (!_root.id) {
      _root.id = 'ui-modal' + generateId();
    }

    return _root;
  }
  static render<T extends new (...args: any[]) => any>(this: T,el: string | HTMLDivElement, id: string, opts?: ConstructorParameters<T>[0]): InstanceType<T>;
  static render<T extends new (...args: any[]) => any>(this: T,el: string | HTMLDivElement, opts?: ConstructorParameters<T>[0]): InstanceType<T>;
  static render(...args: any[]) {
    let [el, id, opts] = args;
    const _el = querySelector(el);
    if (!_el) {
      console.warn('el is not exist', el);
      return;
    }
    let _id: string = '';
    if (typeof id === 'string' && id) {
      // 如果id是字符串
      _id = id;
    } else if (typeof id === 'object') {
      // 如果id是对象
      _id = id.id;
      opts = id;
      id = id.id;
    }
    let _modal: Modal | undefined;
    const modalState = modalStore.getState();
    if (_id) {
      // 如果存在id,则判断是否已经存在该id的modal
      _modal = modalStore.getState().getModal(_id);
    }
    if (!_modal) {
      // 不存在modal,则创建一个modal
      // console.log('create modal', id, opts);
      const newModal = new this({ id, ...opts });
      _modal = newModal;
      modalStore.setState({
        modals: [...modalState.modals, newModal],
      });
    }
    _modal.renderEl(_el);
    return _modal;
  }
  static create<T extends new (...args: any[]) => any>(this:T, opts: ModalOpts):InstanceType<T> {
    let _id = opts.id;
    let _modal: Modal | undefined;
    const modalState = modalStore.getState();
    if (_id) {
      // 如果存在id,则判断是否已经存在该id的modal
      _modal = modalStore.getState().getModal(_id);
    }
    if (!_modal) {
      // 不存在modal,则创建一个modal
      // console.log('create modal', id, opts);
      const newModal = new this({ ...opts, id: _id });
      _modal = newModal;
      modalStore.setState({
        modals: [...modalState.modals, newModal],
      });
    }
    return _modal as InstanceType<T>;
  }
  createMask() {
    const mask = document.createElement('div');
    mask.classList.add('ui-modal-mask');
    mask.dataset.id = this.id;
    elAddCS2(mask, this.maskClassName, this.maskStyle, this.defaultStyle?.defaultMaskStyle);
    return mask;
  }
  renderEl(el: HTMLDivElement) {
    const defaultContentStyle = this.defaultStyle?.defaultContentStyle || {
      position: 'absolute',
      padding: '20px',
      left: '50%',
      top: '20%',
      width: '600px',
      background: '#fff',
      borderRadius: '5px',
      boxShadow: '0 0 10px rgba(0,0,0,.1)',
      transform: 'translate(-50%, -50%)',
      maxHeight: '80vh',
      overflow: 'auto',
    };

    const fragment = document.createDocumentFragment();
    const _modalEl = document.createElement('div');
    _modalEl.classList.add('ui-modal-wrapper');
    _modalEl.id = this.id;
    _modalEl.dataset.mid = this.id;
    if (this.mask) {
      const mask = this.createMask();
      _modalEl.appendChild(mask);
    }
    const modalContent = document.createElement('div');
    modalContent.classList.add('ui-modal-content');
    elAddCS2(modalContent, this.contentClassName, this.contentStyle, defaultContentStyle);
    modalContent.appendChild(el);
    _modalEl.appendChild(modalContent);
    fragment.appendChild(_modalEl);

    this.modalElement = _modalEl;
    this.Element = el;
    this.appendRoot(fragment);

    return this;
  }
  appendRoot(document: DocumentFragment) {
    this.root.appendChild(document);
    // 第一次渲染，open为true，显示弹窗
    this.setOpen(this.open);
  }
  setOpen(open: boolean) {
    this.open = open;
    if (this.destroyOnClose) {
      if (open) {
        this.root.appendChild(this.modalElement);
      } else {
        this.cacheFragment.appendChild(this.modalElement);
      }
      return;
    }
    if (this.hideOnClose) {
      if (open) {
        this.modalElement.classList.remove('ui-modal-close');
      } else {
        this.modalElement.classList.add('ui-modal-close');
      }
    }
  }
  unMount() {
    // 返回渲染的的Element, 然后删除modalElement
    const fragment = document.createDocumentFragment();
    fragment.appendChild(this.Element);
    this.modalElement?.remove();

    const modalState = modalStore.getState();
    modalStore.setState({
      modals: modalState.modals.filter((modal) => modal.id !== this.id),
    });
    this.isUse = false;
    this.cacheFragment = new DocumentFragment();
    return fragment;
  }
  /**
   * 保留，暂时不用
   * // TODO: 研究
   * @param force
   * @param opts
   * @returns
   */
  reRender(force?: boolean, opts?: ModalOpts) {
    if (force) {
      this.modalElement?.remove?.();
      this.modalElement = undefined;
      if (!this.Element) return;
      this.renderEl(this.Element);
    }
  }
  async onMaskClose(e?: any) {
    if (this.maskClose) {
      this.setOpen(false);
      this.onClose?.(e);
    }
  }
  setDefaultStyle(key: keyof DefaultStyle<U>, style: ObjCss) {
    this.defaultStyle[key] = style as any;
  }
}

// modal.render('#abc'||document.querySelector('#abc'));
// modal.unmount();
